
from glue.utils import defer_draw, decorate_all_methods
from glue_qt.viewers.matplotlib.data_viewer import MatplotlibDataViewer
from glue_qt.viewers.image.mouse_mode import RoiClickAndDragMode

from glue_heatmap.qt.layer_style_editor import HeatmapLayerStyleEditor
from glue_heatmap.qt.layer_style_editor_subset import HeatmapLayerSubsetStyleEditor
from glue_heatmap.layer_artist import HeatmapLayerArtist, HeatmapSubsetLayerArtist
from glue_heatmap.qt.options_widget import HeatmapOptionsWidget
from glue_heatmap.state import HeatmapViewerState
from glue_heatmap.viewer import MatplotlibHeatmapMixin

__all__ = ["HeatmapViewer"]


@decorate_all_methods(defer_draw)
class HeatmapViewer(MatplotlibHeatmapMixin, MatplotlibDataViewer):
    LABEL = "Heatmap"
    _default_mouse_mode_cls = RoiClickAndDragMode
    _layer_style_widget_cls = {
        HeatmapLayerArtist: HeatmapLayerStyleEditor,
        HeatmapSubsetLayerArtist: HeatmapLayerSubsetStyleEditor,
    }
    _state_cls = HeatmapViewerState
    _options_cls = HeatmapOptionsWidget

    allow_duplicate_data = False

    # NOTE: _data_artist_cls and _subset_artist_cls are not defined - instead
    #       we override get_data_layer_artist and get_subset_layer_artist for
    #       more advanced logic.

    tools = [
        "select:rectangle",
        "select:xrange",
        "select:yrange",
        "image:contrast_bias",
    ]

    def __init__(self, session, parent=None, state=None):
        MatplotlibDataViewer.__init__(
            self, session, wcs=False, parent=parent, state=state
        )
        MatplotlibHeatmapMixin.setup_callbacks(self)
        self.state.add_callback("cluster", self._update_axes)
        self._wcs_set = False

    def closeEvent(self, *args):
        super().closeEvent(*args)
        if self.axes._composite_image is not None:
            self.axes._composite_image.remove()
            self.axes._composite_image = None



import os
import numpy as np
from qtpy.QtWidgets import QDialog, QListWidgetItem
from qtpy.QtCore import Qt


from echo.qt import autoconnect_callbacks_to_qt
from glue_qt.utils import load_ui
from glue.core import Data, BaseData
from glue.core.state_objects import State
from echo import SelectionCallbackProperty
from glue.viewers.matplotlib.state import DeferredDrawCallbackProperty
from glue.core.data_combo_helper import (
    DataCollectionComboHelper,
    ComponentIDComboHelper,
)
from echo import ChoiceSeparator
from glue_heatmap.coords import HeatmapCoordinates

from glue.core.link_helpers import JoinLink


class ExtractToMatrixState(State):
    data = SelectionCallbackProperty()
    subset = SelectionCallbackProperty()
    component = SelectionCallbackProperty()
    row = SelectionCallbackProperty()
    col_names = DeferredDrawCallbackProperty("columns")

    def __init__(self, data_collection):
        super().__init__()
        self.dc = data_collection
        self.data_helper = DataCollectionComboHelper(self, "data", data_collection)
        self.component_helper = ComponentIDComboHelper(
            self, "component", data_collection=data_collection
        )
        self.row_helper = ComponentIDComboHelper(
            self, "row", data_collection=data_collection
        )
        self.add_callback("data", self._on_data_change)
        self._on_data_change()

    def _on_data_change(self, event=None):
        self.component_helper.set_multiple_data(
            [] if self.data is None else [self.data]
        )
        self.row_helper.set_multiple_data([] if self.data is None else [self.data])

        self._sync_subsets()

    def _sync_subsets(self):
        def display_func(subset):
            if subset is None:
                return "All data (no subsets applied)"
            else:
                return subset.label

        subsets = [None] + list(self.data.subsets)

        ExtractToMatrixState.subset.set_choices(self, subsets)
        ExtractToMatrixState.subset.set_display_func(self, display_func)


class ExtractToMatrixDialog(QDialog):
    def __init__(
        self,
        collect,
        default_data=None,
        default_components=None,
        default_row=None,
        default_col_names=None,
        parent=None,
    ):
        super().__init__(parent=parent)

        self.state = ExtractToMatrixState(collect)

        self.ui = load_ui(
            "extract_to_matrix.ui", self, directory=os.path.dirname(__file__)
        )
        self._connections = autoconnect_callbacks_to_qt(self.state, self.ui)

        if default_data is not None:
            self.state.data = default_data
        if default_components is not None:
            self.default_components = default_components
        if default_row is not None:
            self.state.row = default_row
        if default_col_names is not None:
            self.state.col_names = default_col_names

        self.ui.button_ok.clicked.connect(self.accept)
        self.ui.button_cancel.clicked.connect(self.reject)
        self.ui.button_select_none.clicked.connect(self.select_none)
        self.ui.button_select_all.clicked.connect(self.select_all)

        self.ui.list_component.itemChanged.connect(self._on_check_change)

        self.state.add_callback("component", self._on_data_change)

        self._on_data_change()

    def _on_data_change(self, *event):
        components = getattr(type(self.state), "component").get_choices(self.state)

        self.ui.list_component.clear()

        for component in components:
            if isinstance(component, ChoiceSeparator):
                item = QListWidgetItem(str(component))
                item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
                item.setForeground(Qt.gray)
            else:
                item = QListWidgetItem(component.label)
                if self.default_components is not None:
                    if component in self.default_components:
                        item.setCheckState(Qt.Checked)
                    else:
                        item.setCheckState(Qt.Unchecked)
                else:
                    item.setCheckState(Qt.Checked)
            self.ui.list_component.addItem(item)

    def _on_check_change(self, *event):
        any_checked = False

        for idx in range(self.ui.list_component.count()):
            item = self.ui.list_component.item(idx)
            if item.checkState() == Qt.Checked:
                any_checked = True
                break

        self.button_ok.setEnabled(any_checked)

    def select_none(self, *event):
        self._set_all_checked(False)

    def select_all(self, *event):
        self._set_all_checked(True)

    def _set_all_checked(self, check_state):
        for idx in range(self.ui.list_component.count()):
            item = self.ui.list_component.item(idx)
            item.setCheckState(Qt.Checked if check_state else Qt.Unchecked)

    def _apply(self, do_dialog=True):
        components = []
        for idx in range(self.ui.list_component.count()):
            item = self.ui.list_component.item(idx)
            if item.checkState() == Qt.Checked:
                components.append(self.state.data.id[item.text()])
        if self.state.subset is None:
            data = self.state.data
        else:
            data = self.state.subset
        self.extract_data(
            data,
            self.state.dc,
            components=components,
            rows=self.state.row,
            col_title=self.state.col_names,
        )

    def extract_data(self, data, data_collection, components=[], rows=[], col_title=""):
        column_names = [x.label for x in components]
        row_names = data[rows]
        values = np.vstack([data[comp] for comp in components])
        y_cats = np.array([column_names for x in range(values.shape[1])]).T
        x_cats = np.array([row_names for x in range(values.shape[0])])
        new_data = Data(
            values=values,
            y_cats=y_cats,
            x_cats=x_cats,
            label=f"Matrix from {data.label}",
            coords=HeatmapCoordinates(row_names, column_names, rows.label, col_title),
        )
        data_collection.append(new_data)

        # row_link = LinkSame(new_data.id['x_cats'], data.id[rows] )
        if isinstance(data, BaseData):
            pass
        else:
            data = data.data
        row_link = JoinLink(
            cids1=[new_data.id["x_cats"]],
            cids2=[data.id[rows]],
            data1=new_data,
            data2=data,
        )
        data_collection.add_link(row_link)

    @classmethod
    def create_matrix(
        cls,
        collect,
        default_data=None,
        default_components=None,
        default_row=None,
        default_col_names=None,
        parent=None,
    ):
        self = cls(
            collect,
            parent=parent,
            default_data=default_data,
            default_components=default_components,
            default_row=default_row,
            default_col_names=default_col_names,
        )
        value = self.exec_()

        if value == QDialog.Accepted:
            self._apply()


if __name__ == "__main__":
    from glue.core import DataCollection
    from glue.utils.qt import get_qapp

    data1 = Data(
        genes=["g1", "g2", "g3", "g4"],
        A=[1, 2, 3, 4],
        B=[2, 3, 4, 5],
        C=[4, 5, 6, 7],
        label="data1",
    )
    data2 = Data(a=[1, 2, 3], b=[2, 3, 4], label="data2")

    dc = DataCollection([data1, data2])

    state1 = data1.id["genes"] == "g2"
    state2 = data1.id["genes"] == "g3"
    state = state1 | state2
    subset_group = dc.new_subset_group("g2|g3", state)
    app = get_qapp()

    dialog = ExtractToMatrixDialog(dc, default=data1)
    value = dialog.exec_()
    if value == QDialog.Accepted:
        dialog._apply()



from glue.viewers.image.viewer import MatplotlibImageMixin
from functools import partial
import math

from matplotlib.ticker import FixedLocator, FuncFormatter
from glue.core.util import tick_linker
from glue.core.subset import roi_to_subset_state

from glue_heatmap.layer_artist import HeatmapLayerArtist, HeatmapSubsetLayerArtist

__all__ = ["MatplotlibHeatmapMixin"]


def set_locator(axis_min, axis_max, tick_labels, axis):
    axis_range = math.ceil(axis_max) - math.floor(
        axis_min
    )  # TODO: What is the axes ranges are flipped?
    max_num_cats = min(int(axis_range), 30)
    locator = FixedLocator(
        range(math.floor(axis_min), math.ceil(axis_max), 1), nbins=max_num_cats
    )
    format_func = partial(tick_linker, tick_labels)
    formatter = FuncFormatter(format_func)
    axis.set_major_locator(locator)
    axis.set_major_formatter(formatter)


class MatplotlibHeatmapMixin(MatplotlibImageMixin):
    def limits_from_mpl(self, *args, **kwargs):
        super().limits_from_mpl(*args, **kwargs)

        if self.state.reference_data is None:
            return

        x_ticks = self.state.x_categories
        y_ticks = self.state.y_categories

        set_locator(self.state.x_min, self.state.x_max, x_ticks, self.axes.xaxis)
        set_locator(self.state.y_min, self.state.y_max, y_ticks, self.axes.yaxis)

    def update_x_ticklabel(self, *event):
        # Original image viewer calls this functions assuming
        # we are using a WCSAxes object, which we are not in the Heatmap
        self.axes.tick_params(axis="x", labelsize=self.state.x_ticklabel_size)
        self.axes.xaxis.get_offset_text().set_fontsize(self.state.x_ticklabel_size)
        self.redraw()

    def update_y_ticklabel(self, *event):
        # Original image viewer calls this functions assuming
        # we are using a WCSAxes object, which we are not in the Heatmap
        self.axes.tick_params(axis="y", labelsize=self.state.y_ticklabel_size)
        self.axes.yaxis.get_offset_text().set_fontsize(self.state.y_ticklabel_size)
        self.redraw()

    def _update_axes(self, *args):
        if self.state.x_att_world is not None:
            self.state.x_axislabel = self.state.x_att_world.label
            x_ticks = self.state.x_categories
            set_locator(0, x_ticks.shape[0], x_ticks, self.axes.xaxis)
            # We want to rotate "long" labels and expand the margins
            self.axes.tick_params(axis="x", labelrotation=90)
            self.axes.resizer.margins = [1, 0.1, 1.2, 0.1]

        if self.state.y_att_world is not None:
            self.state.y_axislabel = self.state.y_att_world.label
            y_ticks = self.state.y_categories
            set_locator(0, y_ticks.shape[0], y_ticks, self.axes.yaxis)
            # Expand the margins if the tick labels are "long"
            # self.axes.resizer.margins = [1, 0.1, 1.2, 0.1]

        self.state.reset_limits()
        self.axes.figure.canvas.draw_idle()

    def _set_wcs(self, event=None, relim=True):  # TODO: Do we really need this?
        if (
            self.state.x_att is None
            or self.state.y_att is None
            or self.state.reference_data is None
        ):
            return

        _ = getattr(self.state.reference_data, "coords", None)

        # Reset the axis labels to match the fact that the new axes have no labels
        self.state.x_axislabel = ""
        self.state.y_axislabel = ""

        self._update_appearance_from_settings()
        self._update_axes()

        if relim:
            self.state.reset_limits()

        self._wcs_set = True

    def get_data_layer_artist(self, layer=None, layer_state=None):
        if layer.ndim == 1:
            pass
            # cls = self._scatter_artist
        else:
            cls = HeatmapLayerArtist
        return self.get_layer_artist(cls, layer=layer, layer_state=layer_state)

    def get_subset_layer_artist(self, layer=None, layer_state=None):
        if layer.ndim == 1:
            pass
            # cls = self._scatter_artist
        else:
            cls = HeatmapSubsetLayerArtist
        return self.get_layer_artist(cls, layer=layer, layer_state=layer_state)

    def apply_roi(self, roi, override_mode=None):
        self.redraw()

        if len(self.layers) == 0:
            return

        if (
            self.state.x_att is None
            or self.state.y_att is None
            or self.state.reference_data is None
        ):
            return

        subset_state = roi_to_subset_state(
                roi,
                x_att=self.state.x_att_world,
                x_categories=self.state.x_categories,
                y_att=self.state.y_att_world,
                y_categories=self.state.y_categories)
        self.apply_subset_state(subset_state, override_mode=override_mode)

    def _add_subset(self, subset):
        super()._add_subset(subset)
        self.state._sync_subsets()

    def _remove_subset(self, subset):
        super()._remove_subset(subset)
        self.state._sync_subsets()

    def _update_subset(self, subset):
        super()._update_subset(subset)
        self.state._sync_subsets()
        if subset.subset is self.state.row_subset or subset.subset is self.state.col_subset:
            self.state._calculate_heatmap_data()
            self._update_axes()



from glue.config import menubar_plugin

from glue_heatmap.qt.extract_to_matrix import ExtractToMatrixDialog


@menubar_plugin("Extract colums/rows of a table to a 2D matrix")
def extract_to_matrix_plugin(session, data_collection):
    ExtractToMatrixDialog.create_matrix(data_collection, default=None, parent=None)
    return



import pandas as pd
from glue.core import Data, DataCollection
from glue_qt.viewers.scatter import ScatterViewer
from glue.app.qt import GlueApplication
from glue.core.link_helpers import JoinLink
from glue.core.joins import get_mask_with_key_joins

from glue_heatmap.coords import HeatmapCoordinates
from glue_heatmap.qt import HeatmapViewer
from glue.core.exceptions import IncompatibleAttribute
import traceback


def norecurse(f):
    def func(*args, **kwargs):
        if len([x[2] for x in traceback.extract_stack() if x[2] == f.__name__]) > 0:
            raise IncompatibleAttribute
        return f(*args, **kwargs)
    return func


class MatrixData(Data):
    def get_mask(self, subset_state, view=None):
        try:
            return subset_state.to_mask(self, view=view)
        except IncompatibleAttribute:
            return norecurse(get_mask_with_key_joins)(self, self._key_joins, subset_state, view=view)


def demo():
    qtl_matrix = pd.read_csv("islet_eQTL_matrix.csv").drop_duplicates(
        subset="marker.id"
    )
    strain_names = ["A", "B", "C", "D", "E", "F", "G", "H"]
    values = qtl_matrix[strain_names].values.T
    marker_names = qtl_matrix["marker.id"].values

    heatmap_data = MatrixData(
        values=values,
        label="test",
        coords=HeatmapCoordinates(
            marker_names, strain_names, "Marker Name", "Parent Strain"
        ),
    )
    strains = Data(strain=["A", "B", "C", "D", "E"], label="strains")
    markers = Data(markers=["6_125562888", "9_50632541", "16_26087386", "2_59628224"], label="markers")

    dc = DataCollection([heatmap_data, strains, markers])

    ga = GlueApplication(dc)

    # Using JoinLinks on both rows and cols causes all subsets to propagate through the matrix
    # which is not ideal. In real usage this is probably a DataAnnData object
    marker_link = JoinLink(cids1=[heatmap_data.id["Marker Name"]], cids2=[markers.id["markers"]], data1=heatmap_data, data2=markers)
    # marker_link = LinkSame(heatmap_data.id["Marker Name"], markers.id["markers"])
    dc.add_link(marker_link)

    strain_link = JoinLink(cids1=[heatmap_data.id["Parent Strain"]], cids2=[strains.id["strain"]], data1=heatmap_data, data2=strains)
    # strain_link = LinkSame(heatmap_data.id["Parent Strain"], strains.id["strain"])
    dc.add_link(strain_link)

    t = ga.new_data_viewer(HeatmapViewer)
    t.add_data(dc[0])

    u = ga.new_data_viewer(ScatterViewer)
    u.add_data(dc[1])

    v = ga.new_data_viewer(ScatterViewer)
    v.add_data(dc[2])

    ga.start()


if __name__ == "__main__":
    demo()



"""
There is also a BaseImageLayerArtist
and an ImageSubsetArray.

Do we need to modify either of these?
"""


from glue.viewers.image.layer_artist import ImageLayerArtist, ImageSubsetLayerArtist

from glue_heatmap.state import HeatmapLayerState, HeatmapSubsetLayerState


__all__ = ["HeatmapLayerArtist", "HeatmapSubsetLayerArtist"]


class HeatmapLayerArtist(ImageLayerArtist):
    """ """

    _layer_state_cls = HeatmapLayerState

    def __init__(self, axes, viewer_state, layer_state=None, layer=None):
        super().__init__(axes, viewer_state, layer_state=layer_state, layer=layer)


class HeatmapSubsetLayerArtist(ImageSubsetLayerArtist):
    _layer_state_cls = HeatmapSubsetLayerState

    def __init__(self, axes, viewer_state, layer_state=None, layer=None):
        super().__init__(axes, viewer_state, layer_state=layer_state, layer=layer)



from glue.core import BaseData

from glue.viewers.image.state import (
    ImageViewerState,
    ImageLayerState,
    ImageSubsetLayerState,
    BaseImageLayerState
)
from glue.viewers.matplotlib.state import (DeferredDrawSelectionCallbackProperty as DDSCProperty,
                                           DeferredDrawCallbackProperty as DDCProperty)

from glue.core.data_combo_helper import ComboHelper
from glue.core.exceptions import IncompatibleDataException
from glue.utils import unbroadcast
from glue.core.fixed_resolution_buffer import bounds_for_cache
from glue.core.exceptions import IncompatibleAttribute
from echo import delay_callback
from scipy.cluster import hierarchy
import fastcluster

import numpy as np

__all__ = ["HeatmapViewerState", "HeatmapLayerState", "HeatmapSubsetLayerState"]


class HeatmapViewerState(ImageViewerState):
    """
    A state class that includes all the attributes for a Heatmap Viewer.

    A HeatmapViewer will ultimately be able to display both a 2D matrix
    of values (for a dataset with HeatmapCoordinates) and a heatmap generated
    from 1D tabular data. For now, we have focused on the 2D matrix case.
    """
#    reference_data = DDSCProperty(docstring='The dataset that is used to define the '
#                                             'available x_agg and y_agg components as '
#                                             'well as the subsets to display')
#    rows = DDSCProperty(docstring='The component ID in reference_data that will be '
#                                   'displayed as the rows of the heatmap.')
#    cols = DDSCProperty(docstring='The component ID in reference_data that will be '
#                                     'displayed as the columns of the heatmap.')
#
#    aggregation = DDSCProperty(docstring='The aggregation function to apply'
#                                             'on values to generate the heatmap.'
#                                             'Must be count if values is None.')
#    values = DDSCProperty(docstring='The component ID in reference_data that will be '
#                                     'aggregated over to generate the heatmap. If None, '
#                                     'the count of each row/col combination is used.')
    row_subset = DDSCProperty(docstring='The subset to use for filtering rows')
    col_subset = DDSCProperty(docstring='The subset to use for filtering columns')

    cluster = DDCProperty(False, docstring='Whether to cluster the histogram')

    def __init__(self, **kwargs):
        super().__init__()
        HeatmapViewerState.aspect.set_choices(self, ["auto"])
        self._heatmap_data = None
        self.row_subset_helper = ComboHelper(self, "row_subset")
        self.col_subset_helper = ComboHelper(self, "col_subset")

        def display_func(subset):
            if subset is None:
                return "None"
            else:
                return subset.label

        self.row_subset_helper.choices = [None]
        self.col_subset_helper.choices = [None]
        self.row_subset_helper.display = display_func
        self.col_subset_helper.display = display_func

        self.add_callback('row_subset', self._update_selected_subset, echo_old=True)
        self.add_callback('col_subset', self._update_selected_subset, echo_old=True)
        self.add_callback('cluster', self._do_cluster)
        self._x_categories = []
        self._y_categories = []

        if self.reference_data is not None:
            self._caculate_heatmap_data()

    @property
    def x_categories(self):
        return np.asarray(self._x_categories)

    @property
    def y_categories(self):
        return np.asarray(self._y_categories)

    def _update_selected_subset(self, before=None, after=None, **kwargs):
        """
        Callback for changing the selected subsets for filtering row/col.

        As long as the subset has actually changed, we need to recalculate
        the heatmap data.
        """
        # A callback event for row/col_subset may be triggered if the choices
        # change but the actual selection doesn't - so we avoid doing anything in
        # this case.
        if after is before:
            return
        # We make a subset not visible if it is the one that we use to subset
        # the data, because the subset mask is just the whole data in that
        # case. The user can toggle it back on if they want to.
        for layer in self.layers:
            if not isinstance(layer.layer, BaseData):
                if layer.layer is after:
                    layer.visible = False
                if layer.layer is before:
                    layer.visible = True
        self._calculate_heatmap_data()

    def _calculate_heatmap_data(self, *args, **kwargs):
        """
        Call this to recalculate the heatmap data if reference_data, row_subset, or
        col_subset changes. In the 1D case this will also be called if rows, cols,
        aggregation or values changes.

        If we are recalculating the heatmap, then we disable clustering (since the
        old clustering is not necessarily valie for the new data array and we want
        it to be obvious to the user that they need to recluster).
        """
        self.cluster = False
        self._heatmap_data = self.reference_data[self.reference_data.main_components[0]]
        self.rows_included = np.arange(self._heatmap_data.shape[0])
        self.cols_included = np.arange(self._heatmap_data.shape[1])

        if self.row_subset is not None:
            try:
                unraveled_indices = np.unravel_index(self.row_subset.to_index_list(), self.reference_data.shape)
                self.rows_included = np.unique(unraveled_indices[0])
            except IncompatibleAttribute:
                pass
        if self.col_subset is not None:
            try:
                unraveled_indices = np.unravel_index(self.col_subset.to_index_list(), self.reference_data.shape)
                self.cols_included = np.unique(unraveled_indices[1])
            except IncompatibleAttribute:
                pass

        self._x_categories = self.reference_data.coords.get_tick_labels("x")[self.cols_included]
        self._y_categories = self.reference_data.coords.get_tick_labels("y")[self.rows_included]
        self._heatmap_data = self._heatmap_data[self.rows_included, :][:, self.cols_included]

        self.reset_limits()

    def _set_reference_data(self):
        """
        Set the reference data for the viewer and calculate
        the heatmap and available subsets.
        """
        if self.reference_data is None:
            for layer in self.layers:
                if isinstance(layer.layer, BaseData):
                    self.reference_data = layer.layer
                    self._calculate_heatmap_data()
                    self._sync_subsets()

                    return
        elif self._heatmap_data is None:
            self._calculate_heatmap_data()
            self._sync_subsets()

    def _sync_subsets(self):
        """
        Generate a list of subsets to display as options for filtering.

        Currently, this just takes all subsets in the data collection,
        but in theory we could try to be smart and only show subsets
        that are defined on reference_data and/or show subsets that
        are meaningful for filtering rows/cols (though this is not
        necessarily well-defined).
        """
        if self.reference_data is None:
            return
        # Perhaps we should filter this list of subsets
        # in some "smart" way to be really row/col things?
        subsets = [None] + list(self.reference_data.subsets)
        self.row_subset_helper.choices = subsets
        self.col_subset_helper.choices = subsets

    def reset_limits(self):

        if self._heatmap_data is None or self.x_att is None or self.y_att is None:
            return

        nx = self._heatmap_data.shape[self.x_att.axis]
        ny = self._heatmap_data.shape[self.y_att.axis]

        with delay_callback(self, 'x_min', 'x_max', 'y_min', 'y_max'):
            self.x_min = -0.5
            self.x_max = nx - 0.5
            self.y_min = -0.5
            self.y_max = ny - 0.5
            # We need to adjust the limits in here to avoid triggering all
            # the update events then changing the limits again.
            self._adjust_limits_aspect()

    def _do_cluster(self, *args):
        """
        Apply hierarchical clustering to the heatmap data.

        We preserve the unclustered data and coordinates so that we can
        revert to the original data if the user toggles clustering off.
        """
        if self.cluster:
            data = self._heatmap_data
            orig_xticks = self._x_categories
            orig_yticks = self._y_categories

            self.orig_coords = (orig_xticks, orig_yticks)
            self.orig_data = data

            row_linkage = fastcluster.linkage_vector(data, method='ward', metric='euclidean')
            row_dendro = hierarchy.dendrogram(row_linkage, no_plot=True)
            col_linkage = fastcluster.linkage_vector(data.T, method='ward', metric='euclidean')
            col_dendro = hierarchy.dendrogram(col_linkage, no_plot=True)
            new_row_ind = row_dendro['leaves']
            new_col_ind = col_dendro['leaves']

            new_xticks = [orig_xticks[i] for i in new_col_ind]
            new_yticks = [orig_yticks[i] for i in new_row_ind]

            new_data = data[new_row_ind, :][:, new_col_ind]
            self._x_categories = np.array(new_xticks)
            self._y_categories = np.array(new_yticks)
            self.orig_included = (self.rows_included, self.cols_included)
            self.rows_included = self.rows_included[new_row_ind]
            self.cols_included = self.cols_included[new_col_ind]

            self._heatmap_data = new_data
        else:
            self._x_categories = self.orig_coords[0]
            self._y_categories = self.orig_coords[1]
            self.rows_included = self.orig_included[0]
            self.cols_included = self.orig_included[1]

            self._heatmap_data = self.orig_data


class BaseHeatmapLayerState(BaseImageLayerState):
    """
    To reuse most of the ImageViewer we simply override the get_sliced_data
    method here to return the heatmap data. We can dramatically simplify this
    to remove all the slicing stuff, although the name needs to stay the same.
    """

    def get_sliced_data(self, view=None, bounds=None):
        """
        Override BaseImageLayerState.get_sliced_data to return just the
        underlying heatmap array.
        """
        if self.viewer_state._heatmap_data is None:
            return

        full_view, agg_func, transpose = self.viewer_state.numpy_slice_aggregation_transpose

        x_axis = self.viewer_state.x_att.axis
        y_axis = self.viewer_state.y_att.axis

        # For this method, we make use of Data.compute_fixed_resolution_buffer,
        # which requires us to specify bounds in the form (min, max, nsteps).
        # We also allow view to be passed here (which is a normal Numpy view)
        # and, if given, translate it to bounds. If neither are specified,
        # we behave as if view was [slice(None), slice(None)].

        def slice_to_bound(slc, size):
            min, max, step = slc.indices(size)
            n = (max - min - 1) // step
            max = min + step * n
            return (min, max, n + 1)

        if bounds is None:

            # The view should be that which should just be applied to the data
            # slice, not to all the dimensions of the data - thus it should have at
            # most two dimensions

            if view is None:
                view = [slice(None), slice(None)]
            elif len(view) == 1:
                view = view + [slice(None)]
            elif len(view) > 2:
                raise ValueError('view should have at most two elements')

            full_view[x_axis] = view[1]
            full_view[y_axis] = view[0]

        else:

            full_view[x_axis] = bounds[1]
            full_view[y_axis] = bounds[0]

        for i in range(self.viewer_state.reference_data.ndim):
            if isinstance(full_view[i], slice):
                full_view[i] = slice_to_bound(full_view[i], self.viewer_state.reference_data.shape[i])

        # We now get the fixed resolution buffer

        if isinstance(self.layer, BaseData):
            image = compute_fixed_resolution_buffer(self.viewer_state._heatmap_data, full_view)
        else:
            image = compute_fixed_resolution_buffer(self.viewer_state._heatmap_data,
                                                    full_view, subset_state=self.layer.subset_state,
                                                    ref_state=self.viewer_state)

        # We apply aggregation functions if needed

        if agg_func is None:
            if image.ndim != 2:
                raise IncompatibleDataException()
        else:
            if image.ndim != len(agg_func):
                raise ValueError("Sliced image dimensions ({0}) does not match "
                                 "aggregation function list ({1})"
                                 .format(image.ndim, len(agg_func)))
            for axis in range(image.ndim - 1, -1, -1):
                func = agg_func[axis]
                if func is not None:
                    image = func(image, axis=axis)
            if image.ndim != 2:
                raise ValueError("Image after aggregation should have two dimensions")

        # And finally we transpose the data if the order of x/y is different
        # from the native order.

        if transpose:
            image = image.transpose()

        return image


class HeatmapLayerState(BaseHeatmapLayerState, ImageLayerState):
    """
    A state class that includes all the attributes for data layers in a Heatmap Viewer
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class HeatmapSubsetLayerState(BaseHeatmapLayerState, ImageSubsetLayerState):
    """
    A state class that includes all the attributes for subset layers in Heatmap Viewer
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


# Is there a problem with having this here?
ARRAY_CACHE = {}
PIXEL_CACHE = {}


def compute_fixed_resolution_buffer(data, bounds, subset_state=None, ref_state=None, cache_id=None):
    """
    Get a fixed-resolution buffer for a 2D array.

    This is a replacement for glue.core.fixed_resolution_buffer.compute_fixed_resolution_buffer
    which works on a 2D numpy array and subsets associated with HeatmapViewerState objects.

    Parameters
    ----------
    data : np.ndarray
        The data array to extract a fixed-resolution buffer from.
    bounds : list
        The list of bounds for the fixed resolution buffer. This list should
        have as many items as there are dimensions in ``data``. Each
        item should either be a scalar value, or a tuple of ``(min, max, nsteps)``.
    subset_state : `~glue.core.subset.SubsetState`, optional
        If specified, a subset state to apply to the data before extracting
        the fixed resolution buffer.
    ref_state : State object,  optional
        If specified, a state object that allows us to convert a subset state
        into a mask on the heatmap
    """
    for bound in bounds:
        if isinstance(bound, tuple) and bound[2] < 1:
            raise ValueError(f"Number of steps in bounds should be >=1 but got bound={bound}")

    # If cache_id is specified, we keep a cached version of the resulting array
    # indexed by cache_id as well as a hash formed of the call arguments to this
    # function. We then check if the resulting array already exists in the cache.

    if cache_id is not None:

        # Use uuid for component ID since otherwise component IDs don't return
        # False when comparing two different CIDs (instead they return a subset state).
        # For bounds we use a special wrapper that can identify wildcards.
        current_array_hash = (data, bounds)
        current_pixel_hash = (data)

        if cache_id in ARRAY_CACHE:
            if ARRAY_CACHE[cache_id]['hash'] == current_array_hash:
                return ARRAY_CACHE[cache_id]['array']

        # To save time later, if the pixel cache doesn't match at the level of the
        # data and target_data, we just reset the cache.
        if cache_id in PIXEL_CACHE:
            if PIXEL_CACHE[cache_id]['hash'] != current_pixel_hash:
                PIXEL_CACHE.pop(cache_id)

    # Start off by generating arrays of coordinates in the original dataset
    pixel_coords = [np.linspace(*bound) if isinstance(bound, tuple) else bound for bound in bounds]
    pixel_coords = np.meshgrid(*pixel_coords, indexing='ij', copy=False)

    # Keep track of the original shape of these arrays
    original_shape = pixel_coords[0].shape

    # Now loop through the dimensions of 'data' to find the corresponding
    # coordinates in the frame of view of this dataset.

    translated_coords = []

    invalid_all = np.zeros(original_shape, dtype=bool)

    for ipix in range(data.ndim):

        # At this point, if cache_id is in PIXEL_CACHE, we know that data and
        # target_data match so we just check the bounds. Note that the bounds
        # include the AnyScalar wildcard for any dimensions that don't impact
        # the pixel coordinates here. We do this so that we don't have to
        # recompute the pixel coordinates when e.g. slicing through cubes.

        if cache_id in PIXEL_CACHE and ipix in PIXEL_CACHE[cache_id] and PIXEL_CACHE[cache_id][ipix]['bounds'] == bounds:

            translated_coord = PIXEL_CACHE[cache_id][ipix]['translated_coord']
            dimensions = PIXEL_CACHE[cache_id][ipix]['dimensions']
            invalid = PIXEL_CACHE[cache_id][ipix]['invalid']

        else:

            # The returned coordinates may often be a broadcasted array. To convert
            # the coordinates to integers and check which ones are within bounds, we
            # thus operate on the un-broadcasted array, before broadcasting it back
            # to the original shape.
            translated_coord = np.round(unbroadcast(pixel_coords[ipix])).astype(int)
            invalid = (translated_coord < 0) | (translated_coord >= data.shape[ipix])

            # Since we are going to be using these coordinates later on to index an
            # array, we need the coordinates to be within the array, so we reset
            # any invalid coordinates and keep track of which pixels are invalid
            # to reset them later.
            translated_coord[invalid] = 0

            # We now populate the cache
            if cache_id is not None:

                if cache_id not in PIXEL_CACHE:
                    PIXEL_CACHE[cache_id] = {'hash': current_pixel_hash}

                PIXEL_CACHE[cache_id][ipix] = {'translated_coord': translated_coord,
                                               'dimensions': dimensions,
                                               'invalid': invalid,
                                               'bounds': bounds_for_cache(bounds, dimensions)}

        invalid_all |= invalid

        # Broadcast back to the original shape and add to the list
        translated_coords.append(np.broadcast_to(translated_coord, original_shape))

    translated_coords = tuple(translated_coords)

    if subset_state is None:
        array = data[translated_coords].astype(float)
        invalid_value = -np.inf
    else:
        og_mask = ref_state.reference_data.get_mask(subset_state)  # This is the mask on the original data
        mask_red = og_mask[ref_state.rows_included, :][:, ref_state.cols_included]  # convert from full size to heatmap
        array = mask_red[translated_coords]  # Downsample
        invalid_value = False

    if np.any(invalid_all):
        if not array.flags.writeable:
            array = np.array(array, dtype=type(invalid_value))
        array[invalid_all] = invalid_value

    # Drop dimensions for which bounds were scalars
    slices = []
    for bound in bounds:
        if isinstance(bound, tuple):
            slices.append(slice(None))
        else:
            slices.append(0)

    array = array[tuple(slices)]

    if cache_id is not None:

        # For the bounds, we use a special wildcard for bounds that don't affect
        # the result. This will allow the cache to match regardless of the
        # value for those bounds. However, we only do this for scalar bounds.

        cache_bounds = bounds_for_cache(bounds, [0, 1])

        current_array_hash = current_array_hash[:1] + (cache_bounds,) + current_array_hash[2:]

        ARRAY_CACHE[cache_id] = {'hash': current_array_hash, 'array': array}

    return array



import numpy as np
from glue.core.coordinates import IdentityCoordinates


class HeatmapCoordinates(IdentityCoordinates):
    def __init__(self, x_tick_names, y_tick_names, x_tick_label, y_tick_label, *args, **kwargs):
        super().__init__(n_dim=2)
        self._x_tick_names = np.array(x_tick_names).astype(str)
        self._y_tick_names = np.array(y_tick_names).astype(str)
        self._names = [self._x_tick_names, self._y_tick_names]
        self._x_tick_label = x_tick_label
        self._y_tick_label = y_tick_label

    def get_tick_labels(self, axis_name):
        if axis_name == "x":
            return self._x_tick_names
        elif axis_name == "y":
            return self._y_tick_names

    def pixel_to_world_values(self, *args):
        return tuple([self._names[i][np.round(x)] for i, x in enumerate(args)])

    def world_to_pixel_values(self, *args):
        return tuple([self._names[i].index(x) for i, x in enumerate(args)])

    @property
    def world_axis_names(self):
        # Returns an iterable of strings given the names of the world
        # coordinates for each axis.
        return [self._x_tick_label, self._y_tick_label]

    def __gluestate__(self, context):
        return dict(x_tick_names=context.do(self._x_tick_names),
                    y_tick_names=context.do(self._y_tick_names),
                    x_tick_label=self._x_tick_label,
                    y_tick_label=self._y_tick_label)

    @classmethod
    def __setgluestate__(cls, rec, context):
        return cls(context.object(rec['x_tick_names']),
                   context.object(rec['y_tick_names']),
                   x_tick_label=rec['x_tick_label'],
                   y_tick_label=rec['y_tick_label'])



from .data_viewer import HeatmapViewer  # noqa
#from .subset_tool import SubsetTool  # noqa
#from ..menubar_plugin import extract_to_matrix_plugin  # noqa


def setup():
    from glue_qt.config import qt_client

    qt_client.add(HeatmapViewer)



import os
from qtpy import QtWidgets

from echo.qt import autoconnect_callbacks_to_qt
from glue_qt.utils import load_ui


class HeatmapLayerSubsetStyleEditor(QtWidgets.QWidget):
    def __init__(self, layer, parent=None):
        super().__init__(parent=parent)

        self.ui = load_ui(
            "layer_style_editor_subset.ui", self, directory=os.path.dirname(__file__)
        )

        connect_kwargs = {"alpha": dict(value_range=(0, 1))}

        self._connections = autoconnect_callbacks_to_qt(
            layer.state, self.ui, connect_kwargs
        )



import os

from qtpy import QtWidgets

from echo.qt import autoconnect_callbacks_to_qt
from glue_qt.utils import load_ui, fix_tab_widget_fontsize
from glue_qt.viewers.image.slice_widget import MultiSliceWidgetHelper
from glue.viewers.matplotlib.state import MatplotlibDataViewerState

__all__ = ["HeatmapOptionsWidget"]


class HeatmapOptionsWidget(QtWidgets.QWidget):
    def __init__(self, viewer_state, session, parent=None):
        super().__init__(parent=parent)

        self.ui = load_ui(
            "options_widget.ui", self, directory=os.path.dirname(__file__)
        )

        fix_tab_widget_fontsize(self.ui.tab_widget)

        self._connections = autoconnect_callbacks_to_qt(viewer_state, self.ui)
        self._connections_axes = autoconnect_callbacks_to_qt(
            viewer_state, self.ui.axes_editor.ui
        )
        connect_kwargs = {"alpha": dict(value_range=(0, 1))}
        self._connections_legend = autoconnect_callbacks_to_qt(
            viewer_state.legend, self.ui.legend_editor.ui, connect_kwargs
        )

        self.viewer_state = viewer_state

        # TODO: Excise this
        self.slice_helper = MultiSliceWidgetHelper(
            viewer_state=self.viewer_state, layout=self.ui.layout_slices
        )

        self.session = session
        self.ui.axes_editor.button_apply_all.clicked.connect(self._apply_all_viewers)

    def _apply_all_viewers(self):
        for tab in self.session.application.viewers:
            for viewer in tab:
                if isinstance(viewer.state, MatplotlibDataViewerState):
                    viewer.state.update_axes_settings_from(self.viewer_state)



import os
from qtpy import QtWidgets

from echo.qt import autoconnect_callbacks_to_qt
from glue_qt.utils import load_ui


class HeatmapLayerStyleEditor(QtWidgets.QWidget):
    def __init__(self, layer, parent=None):
        super().__init__(parent=parent)

        self.ui = load_ui(
            "layer_style_editor.ui", self, directory=os.path.dirname(__file__)
        )

        connect_kwargs = {
            "alpha": dict(value_range=(0, 1)),
            "contrast": dict(value_range=(0.1, 10), log=True),
            "bias": dict(value_range=(1.5, -0.5)),
        }

        self._connections = autoconnect_callbacks_to_qt(
            layer.state, self.ui, connect_kwargs
        )

        layer._viewer_state.add_callback("color_mode", self._update_color_mode)

        self._update_color_mode(layer._viewer_state.color_mode)

        self.ui.bool_global_sync.setToolTip(
            "Whether to sync the color and transparency with other viewers"
        )

    def _update_color_mode(self, color_mode):
        if color_mode == "Colormaps":
            self.ui.color_color.hide()
            self.ui.combodata_cmap.show()
        else:
            self.ui.color_color.show()
            self.ui.combodata_cmap.hide()



from glue.config import viewer_tool
from glue.viewers.common.tool import Tool
from qtpy import QtWidgets
from glue.core.state_objects import State
from echo.qt import autoconnect_callbacks_to_qt
from qtpy.QtWidgets import QMessageBox
import numpy as np
from glue_qt.utils import load_ui
from glue.core.data import Data
from glue_heatmap.coords import HeatmapCoordinates
from echo import SelectionCallbackProperty
from glue.core.data_combo_helper import ComboHelper
import os

__all__ = ["SubsetTool", "SubsetToDataState", "SubsetToDataDialog"]


@viewer_tool
class SubsetTool(Tool):
    icon = "glue_spawn"
    tool_id = "heatmap:subset"
    action_text = "Create a new Dataset from a Subset"
    tool_tip = "Create a new Dataset from a Subset"
    shortcut = "Ctrl+L"

    def __init__(self, viewer):
        super().__init__(viewer)
        self._data_collection = viewer._data
        self.data = viewer.state.reference_data

    def activate(self):
        """
        Fired when the toolbar button is activated
        """

        SubsetToDataDialog.make_subset(
            self.viewer.state.reference_data,
            self._data_collection,
            default=None,
            parent=None,
        )

    def close(self):
        if hasattr(self.viewer, "window_closed"):
            self.viewer.window_closed.disconnect(self._do_close)
        self.viewer = None


def dialog(title, text, icon):
    if icon == "warn":
        icon = QMessageBox.Warning
    elif icon == "info":
        icon = QMessageBox.Information
    info = QMessageBox()
    info.setIcon(icon)
    info.setText(title)
    info.setInformativeText(text)
    info.setStandardButtons(info.Ok)
    _ = info.exec_()
    return True


class SubsetToDataState(State):
    subset = (
        SelectionCallbackProperty()
    )  # Required: a subset that can be applied to data

    def __init__(self, data, data_collection):
        super().__init__()

        self.data = data
        self.data_collection = data_collection
        self.subset_helper = ComboHelper(self, "subset")

        def display_func_label(subset_group):
            return subset_group.label

        self.subset_helper.choices = data_collection.subset_groups
        try:
            self.subset_helper.selection = data_collection.subset_groups[0]
        except IndexError:
            pass
        self.subset_helper.display = display_func_label

    def _on_data_change(self, *args, **kwargs):
        pass


def get_extract_method(first_comp, mask):
    mm = np.ma.MaskedArray(first_comp, mask=mask)
    comp_cols = np.ma.compress_cols(mm)
    if comp_cols.size != 0:
        return "comp_cols"
    comp_rows = np.ma.compress_rows(mm)
    if comp_rows.size != 0:
        return "comp_rows"
    return "max_extent"


def clone_subset_into_data_object(subset):
    """
    A helper function to clone a data object

    https://stackoverflow.com/questions/39206986/numpy-get-rectangle-area-just-the-size-of-mask
    """
    new_data = Data()
    old_data = subset.data

    mask = subset.to_mask()

    i, j = np.where(mask)

    first_comp = old_data.main_components[0]
    method = get_extract_method(old_data[first_comp], ~mask)

    if method == "max_extent":
        indices = np.meshgrid(
            np.arange(min(i), max(i) + 1), np.arange(min(j), max(j) + 1), indexing="ij"
        )
        for component in old_data.main_components:
            new_data.add_component(
                old_data[component][tuple(indices)], label=component.label
            )
    elif method == "comp_rows":
        for component in old_data.main_components:
            mm = np.ma.MaskedArray(old_data[component], mask=~mask)
            comp_rows = np.ma.compress_rows(mm)
            new_data.add_component(comp_rows, label=component.label)
    elif method == "comp_cols":
        for component in old_data.main_components:
            mm = np.ma.MaskedArray(old_data[component], mask=~mask)
            comp_cols = np.ma.compress_cols(mm)
            new_data.add_component(comp_cols, label=component.label)

    if old_data.coords:
        new_y_ticks = old_data.coords._y_tick_names[np.unique(i)]
        new_x_ticks = old_data.coords._x_tick_names[np.unique(j)]
        new_data.coords = HeatmapCoordinates(
            new_x_ticks,
            new_y_ticks,
            old_data.coords._x_tick_label,
            old_data.coords._y_tick_label,
        )

    new_data.label = f"{old_data.label} | {subset.label}"
    return new_data


class SubsetToDataDialog(QtWidgets.QDialog):
    def __init__(self, data, collect, default=None, parent=None):
        super().__init__(parent=parent)

        self.state = SubsetToDataState(data, collect)

        self.ui = load_ui(
            "subset_to_data.ui", self, directory=os.path.dirname(__file__)
        )
        self._connections = autoconnect_callbacks_to_qt(self.state, self.ui)

        self._collect = collect

        if default is not None:
            self.state.data = default

        self.ui.button_ok.clicked.connect(self.accept)
        self.ui.button_cancel.clicked.connect(self.reject)

    def _apply(self, do_dialog=True):
        """ """
        if self.state.subset is not None:
            for subset in self.state.subset.subsets:
                if subset.data == self.state.data:
                    target_subset = subset

        results_data = clone_subset_into_data_object(target_subset)
        self._collect.append(results_data)

        if do_dialog:
            _ = dialog(
                "New dataset created",
                f"The Dataset:\n" f"{results_data.label}\n" f"has been created.",
                "info",
            )

    @classmethod
    def make_subset(cls, data, collect, default=None, parent=None):
        self = cls(data, collect, parent=parent, default=default)
        value = self.exec_()

        if value == QtWidgets.QDialog.Accepted:
            self._apply()



from ..data_viewer import HeatmapViewer
from glue.app.qt import GlueApplication
from glue.core.state import GlueUnSerializer
from glue.core.data import Data
import numpy as np

from glue_heatmap.coords import HeatmapCoordinates


class TestHeatmapViewer(object):
    def setup_method(self, method):
        self.app = GlueApplication()
        self.session = self.app.session
        self.hub = self.session.hub
        self.data_collection = self.session.data_collection
        values = np.array([[1, 2, 3], [4, 5, 6]])
        strain_names = ["A", "B"]
        marker_names = ["1", "2", "3"]
        strain_array = np.array([strain_names for x in range(values.shape[1])]).T
        marker_array = np.array([marker_names for x in range(values.shape[0])])
        self.data = Data(
            values=values,
            x_cats=marker_array,
            y_cats=strain_array,
            coords=HeatmapCoordinates(
                marker_names, strain_names, "Marker Name", "Parent Strain"
            ),
            label="test_data",
        )
        self.data_collection.append(self.data)
        self.viewer = self.app.new_data_viewer(HeatmapViewer)
        self.viewer.add_data(self.data)

    def test_basic(self):
        assert self.viewer is not None
        assert self.viewer.state.x_att_world is self.data.id["Marker Name"]
        assert self.viewer.state.y_att_world is self.data.id["Parent Strain"]
        assert self.viewer.state.cluster is False
        assert self.viewer.state.row_subset is None
        assert self.viewer.state.col_subset is None

    def test_save_and_restore(self, tmpdir):

        filename = tmpdir.join("test_heatmap_session.glu").strpath

        self.session.application.save_session(filename)

        with open(filename, "r") as f:
            session = f.read()

        state = GlueUnSerializer.loads(session)

        ga = state.object("__main__")

        viewer = ga.viewers[0][0]
        assert viewer.state.x_att_world.label == "Marker Name"
        ga.close()
